from . import api_bp
from ...libraries.subnet_scan import ScanConfig
from .. import scan_manager

from flask import request, jsonify
import traceback

# Subnet Scanner API
############################################
@api_bp.route('/api/scan', methods=['POST'])
@api_bp.route('/api/scan/threaded', methods=['POST'])
def scan_subnet_threaded():
    try:
        config = get_scan_config()
        scan = scan_manager.new_scan(config)

        return jsonify({'status': 'running', 'scan_id': scan.uid})
    except:
        return jsonify({'status': 'error', 'traceback': traceback.format_exc()}), 500
    

@api_bp.route('/api/scan/async', methods=['POST'])
def scan_subnet_async():
    config = get_scan_config()
    scan = scan_manager.new_scan(config)
    scan_manager.wait_until_complete(scan.uid)

    return jsonify({'status': 'complete', 'scan_id': scan.uid})

@api_bp.route('/api/scan/<scan_id>', methods=['GET'])
def get_scan(scan_id):
    scan = scan_manager.get_scan(scan_id)
    return jsonify(scan.results.export())

def get_scan_config():
    """
    pulls config from the request body
    """
    data = request.get_json()
    return ScanConfig(
        subnet = data['subnet'],
        port_list= data['port_list'],
        parallelism=data.get('parallelism',1.0)
    )


import os
import json
import uuid
import logging
import ipaddress
import traceback
import threading
from time import time
from time import sleep
from typing import List
from dataclasses import dataclass
from concurrent.futures import ThreadPoolExecutor

from .net_tools import Device
from .decorators import job_tracker
from .ip_parser import parse_ip_input
from .port_manager import PortManager

@dataclass
class ScanConfig:
    subnet: str
    port_list: str
    parallelism: float = 1.0



class SubnetScanner:
    def __init__(
            self, 
            config: ScanConfig
        ):
        self.subnet = parse_ip_input(config.subnet)
        self.port_list = config.port_list
        self.ports: list = PortManager().get_port_list(config.port_list).keys()
        self.running = False
        self.parallelism: float = float(config.parallelism)
        self.subnet_str = config.subnet
        self.job_stats = {'running': {}, 'finished': {}, 'timing': {}}
        self.uid = str(uuid.uuid4())
        self.results = ScannerResults(self)
        self.log: logging.Logger = logging.getLogger('SubnetScanner')
        self.log.debug(f'Instantiated with uid: {self.uid}')
        self.log.debug(f'Port Count: {len(self.ports)} | Device Count: {len(self.subnet)}')


    
    
    def start(self):
        """
        Scan the subnet for devices and open ports.
        """
        self._set_stage('scanning devices')
        self.running = True
        with ThreadPoolExecutor(max_workers=self._t_cnt(256)) as executor:
            futures = {executor.submit(self._get_host_details, str(ip)): str(ip) for ip in self.subnet}
            for future in futures:
                ip = futures[future]
                try:
                    future.result()
                except Exception as e:
                    self.log.error(f'[{ip}] scan failed. details below:\n{traceback.format_exc()}')
                    self.results.errors.append({
                        'basic': f"Error scanning IP {ip}: {e}",
                        'traceback': traceback.format_exc(),
                    })
                
        
        self._set_stage('testing ports')
        self._scan_network_ports()

        self._set_stage('complete')
        self.running = False
        return self.results
        

    def debug_active_scan(self):
        """
            Run this after running scan_subnet_threaded 
            to see the progress of the scan
        """
        while self.running:
            os.system('cls' if os.name == 'nt' else 'clear')
            print(f'{self.uid} - {self.subnet_str}')
            print(f"Scanned: {self.results.devices_scanned}/{self.results.devices_total}")
            print(f"Alive: {len(self.results.devices)}")
            print(f"Running jobs:  {self.job_stats['running']}")
            print(f"Finished jobs: {self.job_stats['finished']}")
            print(f"Job timing:    {self.job_stats['timing']}")
            sleep(1)


    
    def _get_host_details(self, host: str):
        """
        Get the MAC address and open ports of the given host.
        """
        device = Device(host)
        is_alive = self._ping(device)
        self.results.scanned()
        if not is_alive:
            return None
        self.log.debug(f'[{host}] is alive, getting metadata')
        
        device.get_metadata()
        self.results.devices.append(device)
        return True
        
    def _scan_network_ports(self):
        with ThreadPoolExecutor(max_workers=self._t_cnt(10)) as executor:
            futures = {executor.submit(self._scan_ports, device): device for device in self.results.devices}
            for future in futures:
                future.result()

    @job_tracker
    def _scan_ports(self, device: Device):
        self.log.debug(f'[{device.ip}] Initiating port scan')
        device.stage = 'scanning'
        with ThreadPoolExecutor(max_workers=self._t_cnt(128)) as executor:
            futures = {executor.submit(self._test_port, device, int(port)): port for port in self.ports}
            for future in futures:
                future.result()
        self.log.debug(f'[{device.ip}] Completed port scan')
        device.stage = 'complete'
    
    @job_tracker
    def _test_port(self,host: Device, port: int):
        """
        Test if a port is open on a given host.
        Device class handles tracking open ports.
        """
        return host.test_port(port)
    
        

    @job_tracker
    def _ping(self, host: Device):
        """
        Ping the given host and return True if it's reachable, False otherwise.
        """
        return host.is_alive(host.ip)
    
    def _t_cnt(self, base_threads: int) -> int:
        """
        Calculate the number of threads to use based on the base number 
        of threads and the parallelism factor.
        """
        return int(base_threads * self.parallelism)
    
    def _set_stage(self,stage):
        self.log.debug(f'[{self.uid}] Moving to Stage: {stage}')
        self.results.stage = stage
    
class ScannerResults:
    def __init__(self,scan: SubnetScanner):
        self.scan = scan
        self.port_list: str = scan.port_list
        self.subnet: str = scan.subnet_str
        self.parallelism: float = scan.parallelism
        self.uid = scan.uid

        self.devices_total: int = len(list(scan.subnet))
        self.devices_alive: int = 0
        self.devices_scanned: int = 0
        self.devices: List[Device] = []

        self.errors: List[str] = []        
        self.running: bool = False
        self.start_time: float = time()
        self.run_time: int = 0
        self.stage = 'instantiated'

        self.log = logging.getLogger('ScannerResults')
        self.log.debug(f'Instantiated Logger For Scan: {self.scan.uid}')


    def scanned(self):
        self.devices_scanned += 1
        

    
    def export(self,out_type=dict) -> str | dict:
        """
            Returns json representation of the scan
        """

        self.running = self.scan.running
        self.run_time = int(round(time() - self.start_time,0))
        self.devices_alive = len(self.devices)

        out = vars(self).copy()
        out.pop('scan')
        out.pop('log')
        
        devices: Device = out.pop('devices')
        sortedDevices = sorted(devices, key=lambda obj: ipaddress.IPv4Address(obj.ip))
        out['devices'] = [vars(device).copy() for device in sortedDevices]
        for device in out['devices']: device.pop('log') 

        if out_type == str:
            return json.dumps(out, indent=2)
        # otherwise return dict
        return out


class ScanManager:
    """
    Maintain active and completed scans in memory for 
    future reference
    """
    def __init__(self):
        self.scans: List[SubnetScanner] = []

    def new_scan(self, config: ScanConfig) -> SubnetScanner:
        scan = SubnetScanner(config)
        self._start(scan)
        self.scans.append(scan)
        return scan

    def get_scan(self,scan_id:str) -> SubnetScanner:
        """
        Get scan by scan.uid
        """
        for scan in self.scans:
            if scan.uid == scan_id:
                return scan

    def wait_until_complete(self,scan_id:str) -> SubnetScanner:
        scan = self.get_scan(scan_id)
        while scan.running:
            sleep(.5)
        return scan

    def _start(self,scan:SubnetScanner):
        t = threading.Thread(target=scan.start)
        t.start()
        return t
    


    


from flask import render_template, request
from . import web_bp
from ...libraries.subnet_scan import SubnetScanner
from ...libraries.net_tools import (
    get_all_network_subnets, 
    smart_select_primary_subnet
)
from .. import scan_manager
import os

# Template Renderer
############################################
@web_bp.route('/', methods=['GET'])
def index():
    subnets = get_all_network_subnets()
    subnet = smart_select_primary_subnet(subnets)
    
    port_list = 'medium'
    parallelism = 0.7
    if scan_id := request.args.get('scan_id'): 
        if scanner := scan_manager.get_scan(scan_id):
            scan = scanner.results.export()
            subnet = scan['subnet']
            port_list = scan['port_list']
            parallelism = scan['parallelism']
    return render_template(
        'main.html',
        subnet=subnet, 
        port_list=port_list, 
        parallelism=parallelism,
        alternate_subnets=subnets
    )

@web_bp.route('/scan/<scan_id>', methods=['GET'])
@web_bp.route('/scan/<scan_id>/<section>', methods=['GET'])
def render_scan(scan_id, section='all'):
    scanner = scan_manager.get_scan(scan_id)
    data = scanner.results.export()
    filter = request.args.get('filter')
    return render_template('scan.html', data=data, section=section, filter=filter)

@web_bp.route('/errors/<scan_id>')
def view_errors(scan_id):
    scanner = scan_manager.get_scan(scan_id)
    data = scanner.results.export()
    return render_template('scan/scan-error.html',data=data)

@web_bp.route('/export/<scan_id>')
def export_scan(scan_id):
    scanner = scan_manager.get_scan(scan_id)
    export_json = scanner.results.export(str)
    return render_template(
        'scan/export.html',
        scan=scanner,
        export_json=export_json
    )

@web_bp.route('/shutdown-ui')
def shutdown_ui():
    return render_template('shutdown.html')

@web_bp.route('/info')
def app_info():
    return render_template('info.html')


import json
import os
from typing import List
from pathlib import Path
from .resource_manager import ResourceManager

PORT_DIR = 'ports'

class PortManager:
    def __init__(self):
        Path(PORT_DIR).mkdir(parents=True, exist_ok=True)
        self.rm = ResourceManager(PORT_DIR)

    def get_port_lists(self) -> List[str]:
        return [f.replace('.json','') for f in self.rm.list() if f.endswith('.json')]
    
    def get_port_list(self, port_list: str) -> dict:
        
        if port_list not in self.get_port_lists(): 
            raise ValueError(f"Port list '{port_list}' does not exist. Available port lists: {self.get_port_lists()}")

        
        data = json.loads(self.rm.get(f'{port_list}.json'))

        return data if self.validate_port_data(data) else None
        
    def create_port_list(self, port_list: str, data: dict) -> bool:
        if port_list in self.get_port_lists(): return False
        if not self.validate_port_data(data): return False

        self.rm.create(f'{port_list}.json', json.dumps(data, indent=2))

        return True
    
    def update_port_list(self, port_list: str, data: dict) -> bool:
        if port_list not in self.get_port_lists(): return False
        if not self.validate_port_data(data): return False

        self.rm.update(f'{port_list}.json', json.dumps(data, indent=2))
        
        return True
    
    def delete_port_list(self, port_list: str) -> bool:
        if port_list not in self.get_port_lists(): return False

        self.rm.delete(f'{port_list}.json')

        return True

    def validate_port_data(self, port_data: dict) -> bool:
        try:
            for port, service in port_data.items():
                port = int(port) # throws if not int
                if not isinstance(service, str): return False

                if not 0 <= port <= 65535: return False
            return True
        except:
            return False
        



import re
import psutil
import socket
import struct
import logging
import platform
import ipaddress
import traceback
import subprocess
from time import sleep
from typing import List
from scapy.all import ARP, Ether, srp

from .mac_lookup import lookup_mac, get_mac
from .ip_parser import get_address_count, MAX_IPS_ALLOWED

log = logging.getLogger('NetTools')


class IPAlive:
    
    def is_alive(self,ip:str) -> bool:
        try:
            self.alive = self._arp_lookup(ip)
        except:
            self.log.debug('failed ARP, falling back to ping')
            self.alive = self._ping_lookup(ip)

        return self.alive

    def _arp_lookup(self,ip,timeout=4):
        arp_request = ARP(pdst=ip)
        broadcast = Ether(dst="ff:ff:ff:ff:ff:ff")
        arp_request_broadcast = broadcast / arp_request

        # Send the packet and receive the response
        answered, _ = srp(arp_request_broadcast, timeout=timeout, verbose=False)

        for sent, received in answered:
            if received.psrc == ip:
                return True
        return False

    def _ping_lookup(self,host, retries=1, retry_delay=1, ping_count=2, timeout=2):
            """
            Ping the given host and return True if it's reachable, False otherwise.
            """
            os = platform.system().lower()
            if os == "windows":
                ping_command = ['ping', '-n', str(ping_count), '-w', str(timeout*1000)]  
            else:
                ping_command = ['ping', '-c', str(ping_count), '-W', str(timeout)]
                
            for _ in range(retries):
                try:
                    output = subprocess.check_output(ping_command + [host], stderr=subprocess.STDOUT, universal_newlines=True)
                    # Check if 'TTL' or 'time' is in the output to determine success
                    if 'TTL' in output.upper():
                        return True
                except subprocess.CalledProcessError:
                    pass  # Ping failed
                sleep(retry_delay)
            return False

class Device(IPAlive):
    def __init__(self,ip:str):
        self.ip: str = ip
        self.alive: bool = None
        self.hostname: str = None
        self.mac_addr: str = None
        self.manufacturer: str = None
        self.ports: List[int] = []
        self.stage: str = 'found'
        self.log = logging.getLogger('Device')

    def get_metadata(self):
        if self.alive:
            self.hostname = self._get_hostname()
            self.mac_addr = self._get_mac_address()
            self.manufacturer = self._get_manufacturer()
            
    
    def test_port(self,port:int) -> bool:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(1)
        result = sock.connect_ex((self.ip, port))
        sock.close()
        if result == 0:
            self.ports.append(port)
            return True
        return False

    def _get_mac_address(self):
        """
        Get the MAC address of a network device given its IP address.
        """
        return get_mac(self.ip)
        
    def _get_hostname(self):
        """
        Get the hostname of a network device given its IP address.
        """
        try:
            hostname = socket.gethostbyaddr(self.ip)[0]
            return hostname
        except socket.herror:
            return None
        
    def _get_manufacturer(self):
        """
        Get the manufacturer of a network device given its MAC address.
        """
        return lookup_mac(self.mac_addr) if self.mac_addr else None
    

def get_ip_address(interface: str):
    """
    Get the IP address of a network interface on Windows or Linux.
    """
    def linux():
        try:
            import fcntl
            import struct
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            ip_address = socket.inet_ntoa(fcntl.ioctl(
                sock.fileno(),
                0x8915,  # SIOCGIFADDR
                struct.pack('256s', interface[:15].encode('utf-8'))
            )[20:24])
            return ip_address
        except IOError:
            return None

    def windows():
        # Get network interfaces and IP addresses using psutil
        net_if_addrs = psutil.net_if_addrs()
        if interface in net_if_addrs:
            for addr in net_if_addrs[interface]:
                if addr.family == socket.AF_INET:  # Check for IPv4
                    return addr.address
        return None

    # Call the appropriate function based on the platform
    if psutil.WINDOWS:
        return windows()
    elif psutil.LINUX:
        return linux()
    else:
        return None

def get_netmask(interface: str):
    """
    Get the netmask of a network interface.
    """
    
    def linux():
        try:
            import fcntl
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            netmask = socket.inet_ntoa(fcntl.ioctl(
                sock.fileno(),
                0x891b,  # SIOCGIFNETMASK
                struct.pack('256s', interface[:15].encode('utf-8'))
            )[20:24])
            return netmask
        except IOError:
            return None

    def windows():
        output = subprocess.check_output("ipconfig", shell=True).decode()
        # Use a regular expression to match both interface and subnet mask
        interface_section_pattern = rf"{interface}.*?Subnet Mask.*?:\s+(\d+\.\d+\.\d+\.\d+)"
        match = re.search(interface_section_pattern, output, re.S)  # Use re.S to allow dot to match newline
        if match:
            return match.group(1)
        return None
    
    if psutil.WINDOWS:
        return windows()
    return linux()

def get_cidr_from_netmask(netmask: str):
    """
    Get the CIDR notation of a netmask.
    """
    binary_str = ''.join([bin(int(x)).lstrip('0b').zfill(8) for x in netmask.split('.')])
    return str(len(binary_str.rstrip('0')))

def get_primary_interface():
    """
    Get the primary network interface based on the default gateway.
    """
    # Get the default gateway information
    gateways = psutil.net_if_addrs()
    default_gw = psutil.net_if_stats()

    # Iterate over the default gateways
    for interface, addrs in gateways.items():
        if default_gw[interface].isup:  # Ensure the interface is up
            for addr in addrs:
                if addr.family == socket.AF_INET:  # Look for IPv4 addresses
                    return interface

    return None

def get_host_ip_mask(ip_with_cidr: str):
    """
    Get the IP address and netmask of a network interface.
    """
    cidr = ip_with_cidr.split('/')[1]
    network = ipaddress.ip_network(ip_with_cidr, strict=False)
    return f'{network.network_address}/{cidr}'

def get_network_subnet(interface = get_primary_interface()):
    """
    Get the network interface and subnet.
    Default is primary interface
    """ 
    try:
        ip_address = get_ip_address(interface)
        netmask = get_netmask(interface)
        # is valid interface?
        if ip_address and netmask:
            cidr = get_cidr_from_netmask(netmask)

            ip_mask = f'{ip_address}/{cidr}'

            return get_host_ip_mask(ip_mask)
    except:
        log.info(f'Unable to parse subnet for interface: {interface}')
        log.debug(traceback.format_exc())
    return

def get_all_network_subnets():
    """
    Get the primary network interface.
    """
    addrs = psutil.net_if_addrs()
    gateways = psutil.net_if_stats()
    subnets = []
    
    for interface, snicaddrs in addrs.items():
        for snicaddr in snicaddrs:
            if snicaddr.family == socket.AF_INET and gateways[interface].isup:
                subnet = get_network_subnet(interface)
                if subnet: 
                    subnets.append({ 
                        'subnet': subnet, 
                        'address_cnt': get_address_count(subnet) 
                    })

    return subnets

def smart_select_primary_subnet(subnets: List[dict]) -> str:
    """
     Finds the largest subnet within max ip range
     not perfect, but works better than subnets[0]
    """
    selected = {}
    for subnet in subnets:
        if selected.get('address_cnt',0) < subnet['address_cnt'] < MAX_IPS_ALLOWED:
            selected = subnet
    if not selected and len(subnets):
        selected = subnets[0]
    return selected['subnet']